#!/bin/sh

#TODO: requires mktemp from coreutils
GOOD=`mktemp`

GOOD_SECTORS=452767
GOOD_SECTORSIZE=512

GOOD_REF="/tmp/ref-good.img"
PAT_REF="/mnt/scratch/ref-pat.img"

DC3DD="src/dc3dd"

TEST_IMG="/tmp/test2.img"
TEST_LOG="/tmp/test2.log"

# no config below this line
GOOD_BYTES=`expr $GOOD_SECTORS \* $GOOD_SECTORSIZE`

# create a "good" image if necessary
function init()
{
    if [ ! -e "$GOOD_REF" ]
    then
        dd if=/dev/urandom of="$GOOD_REF" bs="$GOOD_SECTORSIZE" count="$GOOD_SECTORS"
    else
        #TODO: requires stat from coreutils
        if [ `stat -c %s "$GOOD_REF"` -ne $GOOD_BYTES ]
        then
            dd if=/dev/urandom of="$GOOD_REF" bs="$GOOD_SECTORSIZE" count="$GOOD_SECTORS"
        fi
    fi
}

# clean up temp files from any previous runs
function cleanup()
{
    rm "$TEST_IMG" "$TEST_LOG" "$TEST_IMG".* "$TEST_LOG".*
    if [ -e "$TEST_IMG" -o -e "$TEST_LOG" ]
    then
        echo "Failed to remove test output files"
        exit 2
    fi
}

function good()
{
cleanup

# load reference image into good drive
dd if="$GOOD_REF" of="$GOOD" bs="$GOOD_SECTORSIZE"
if [ $? -ne 0 ]
then
    echo "Failed to load good reference image $GOOD_REF to good device $GOOD"
    exit 1
fi

# get image from good drive
"$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" hash=md5,sha1 sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "Run Failed"
    exit 3
fi

# check the retrieved image

# a. file sizes
#TODO: requires stat from coreutils
REF_BYTES=`stat -c %s "$GOOD_REF"`
IMG_BYTES=`stat -c %s "$TEST_IMG"`

if [ $REF_BYTES -ne $IMG_BYTES ]
then
    echo "Size mismatch"
    exit 3
fi

# b. checksums
REF_SUM=`md5sum "$GOOD_REF" | cut -f 1 -d \ `
IMG_SUM=`md5sum "$TEST_IMG" | cut -f 1 -d \ `
LOG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `

if [ "$REF_SUM" != "$IMG_SUM" ]
then
    echo "Ref/Img Checksum mismatch"
    exit 3
fi

if [ "$REF_SUM" != "$LOG_SUM" ]
then
    echo "Ref/Log Checksum mismatch"
    exit 3
fi

# sector probe/count

LOG_SECTORMODE=`grep sector\ size\: "$TEST_LOG" | awk '{print $4}'`
LOG_SECTORSIZE=`grep sector\ size\: "$TEST_LOG" | awk '{print $3}'`
LOG_SECTORS_FULL=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f1 -d+`
LOG_SECTORS_PART=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f2 -d+`
LOG_BYTES=`grep bytes "$TEST_LOG" | awk '{print $1}'`

# TODO should this be left in?
if [ "$LOG_SECTORSIZE" != "$GOOD_SECTORSIZE" ]
then
    echo "Wrong sector size detected"
    exit 3
fi

if [ "$LOG_SECTORS_PART" != "0" ]
then
    echo "Partial sectors must be zero"
    exit 3
fi

if [ "$LOG_SECTORS_FULL" != "$GOOD_SECTORS" ]
then
    echo "Wrong sector count"
    exit 3
fi

if [ "$LOG_BYTES" != "$GOOD_BYTES" ]
then
    echo "Wrong byte count - log doesn't match size calculated from kernel sectors and sectorsize"
    exit 3
fi

if [ "$LOG_BYTES" != "$IMG_BYTES" ]
then
    echo "Log file bytes doesn't match output file size"
    exit 3
fi

if [ "$LOG_BYTES" != "$REF_BYTES" ]
then
    echo "Log file bytes doesn't match reference file size"
    exit 3
fi

MSGCOUNT=`egrep -c bytes.+copied "$TEST_LOG"`
if [ $MSGCOUNT -ne 1 ]
then
    echo "Log is missing 'bytes copied' message"
    exit 3
fi

# test verify mode

# a. drive to reference file
"$DC3DD" if="$GOOD" vf="$GOOD_REF" conv=noerror,sync log="$TEST_LOG" sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "Verify to Reference Image Failed"
    exit 3
fi

MSGCOUNT=`egrep -c bytes.+compared "$TEST_LOG"`
if [ $MSGCOUNT -ne 1 ]
then
    echo "Log is missing 'bytes compared' message"
    exit 3
fi

MSGCOUNT=`grep -c Verify\ PASSED "$TEST_LOG"`
if [ $MSGCOUNT -ne 1 ]
then
    echo "Log is missing 'Verify PASSED' message"
    exit 3
fi


# b. drive to produced image file
"$DC3DD" if="$GOOD" vf="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "Verify to Created Image Failed"
    exit 3
fi

MSGCOUNT=`egrep -c bytes.+compared "$TEST_LOG"`
if [ $MSGCOUNT -ne 2 ]
then
    echo "Log is missing 'bytes compared' message"
    exit 3
fi

MSGCOUNT=`grep -c Verify\ PASSED "$TEST_LOG"`
if [ $MSGCOUNT -ne 2 ]
then
    echo "Log is missing 'Verify PASSED' message"
    exit 3
fi


# test splitting
cleanup

"$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" splitformat=000 split=100M hash=md5,sha1 sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "Image to split output failed"
    exit 3
fi

IMG_SUM=`cat "$TEST_IMG".* | md5sum | cut -f 1 -d \ `
LOG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `

if [ "$REF_SUM" != "$LOG_SUM" ]
then
    echo "Split: Ref/Log Checksum mismatch"
    exit 3
fi

if [ "$IMG_SUM" != "$REF_SUM" ]
then
    echo "Split: Img/Ref Checksum mismatch"
    exit 3
fi

LOG_SECTORMODE=`grep sector\ size\: "$TEST_LOG" | awk '{print $4}'`
LOG_SECTORSIZE=`grep sector\ size\: "$TEST_LOG" | awk '{print $3}'`
LOG_SECTORS_FULL=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f1 -d+`
LOG_SECTORS_PART=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f2 -d+`
LOG_BYTES=`grep bytes "$TEST_LOG" | awk '{print $1}'`

# TODO should this be left in?
if [ "$LOG_SECTORSIZE" != "$GOOD_SECTORSIZE" ]
then
    echo "Split: Wrong sector size detected"
    exit 3
fi

if [ "$LOG_SECTORS_PART" != "0" ]
then
    echo "Split: Partial sectors must be zero"
    exit 3
fi

if [ "$LOG_SECTORS_FULL" != "$GOOD_SECTORS" ]
then
    echo "Split: Wrong sector count"
    exit 3
fi

if [ "$LOG_BYTES" != "$GOOD_BYTES" ]
then
    echo "Split: Wrong byte count - log doesn't match size calculated from kernel sectors and sectorsize"
    exit 3
fi

IMG_BYTES=0
for a in "$TEST_IMG".*
do
    #TODO: requires stat from coreutils
    SPLIT_BYTES=`stat -c %s "$a"`
    IMG_BYTES=`expr $IMG_BYTES + $SPLIT_BYTES`
done

if [ "$IMG_BYTES" != "$REF_BYTES" ]
then
    echo "Split: image bytes doesn't match reference file size"
    exit 3
fi

if [ "$LOG_BYTES" != "$IMG_BYTES" ]
then
    echo "Split: Log file bytes doesn't match output file size"
    exit 3
fi

# test joining
"$DC3DD" if="$GOOD" vfjoin="$TEST_IMG.000" conv=noerror,sync sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "vfjoin failed"
    exit 3
fi

"$DC3DD" ifjoin="$TEST_IMG.000" vf="$GOOD_REF" conv=noerror,sync sizeprobe=on progress=on
if [ $? -ne 0 ]
then
    echo "ifjoin failed"
    exit 3
fi

# test interrupting
# NOTE input file/device needs to be big enough that the acquisition won't finish
# before the sleep timeout expires

cleanup

set -m
"$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" hash=md5,sha1 sizeprobe=on progress=on& pid=$!;
sleep 1; kill -s 2 %1
wait %1
set +m

#if [ $? -ne 0 ]
#then
#    echo "Run Failed"
#    exit 3
#fi

#TODO: requires stat from coreutils
IMG_BYTES=`stat -c %s "$TEST_IMG"`
LOG_BYTES=`grep bytes "$TEST_LOG" | awk '{print $1}'`

if [ "$LOG_BYTES" != "$IMG_BYTES" ]
then
    echo "Interrupt: Log file bytes doesn't match output file size"
    exit 3
fi

IMG_SUM=`md5sum "$TEST_IMG" | cut -f 1 -d \ `
LOG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `

if [ "$IMG_SUM" != "$LOG_SUM" ]
then
    echo "Interrupt: Img/Log Checksum mismatch"
    exit 3
fi

}

function zero
{
    cleanup

    ZERO_COUNT=$1
    DEF_SECTORSIZE=512

    "$DC3DD" if="/dev/zero" of="/dev/null" log="$TEST_LOG" count="$ZERO_COUNT" sizeprobe=on progress=on
    if [ $? -ne 0 ]
    then
        echo "Imaging from /dev/zero"
        exit 5
    fi

    LOG_SECTORMODE=`grep sector\ size\: "$TEST_LOG" | awk '{print $4}'`
    LOG_SECTORSIZE=`grep sector\ size\: "$TEST_LOG" | awk '{print $3}'`
    LOG_SECTORS_FULL=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f1 -d+`
    LOG_SECTORS_PART=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f2 -d+`
    LOG_SECTORS_OUT_FULL=`grep sectors\ out "$TEST_LOG" | awk '{print $1}' | cut -f1 -d+`
    LOG_SECTORS_OUT_PART=`grep sectors\ out "$TEST_LOG" | awk '{print $1}' | cut -f2 -d+`
    LOG_BYTES=`grep bytes "$TEST_LOG" | awk '{print $1}'`

    if [ "$LOG_SECTORMODE" != "(assumed)" ]
    then
        echo "Split: should be assuming sector size for zero/pattern fill"
        exit 5
    fi

    if [ "$LOG_SECTORSIZE" != "$DEF_SECTORSIZE" ]
    then
        echo "Split: Wrong sector size assumed"
        exit 5
    fi

    if [ "$LOG_SECTORS_PART" != "0" ]
    then
        echo "Split: Partial (error) sectors in must be 0"
        exit 5
    fi

    if [ "$LOG_SECTORS_FULL" != "$ZERO_COUNT" ]
    then
        echo "Split: Wrong sectors in full count"
        exit 5
    fi

    if [ "$LOG_SECTORS_OUT_PART" != "0" ]
    then
        echo "Split: Partial (error) sectors out must be 0"
        exit 5
    fi

    if [ "$LOG_SECTORS_OUT_FULL" != "$ZERO_COUNT" ]
    then
        echo "Split: Full sectors out count wrong"
        exit 5
    fi

    if [ "$LOG_BYTES" != `expr $ZERO_COUNT \* $DEF_SECTORSIZE` ]
    then
        echo "Split: Wrong byte count - log doesn't match size calculated from kernel sectors and sectorsize"
        exit 5
    fi
}

function zeros
{
    zero       1 
    zero      10 
    zero     500 
    zero     512 
    zero    1000 
    zero    1024 
    zero 1000000
    zero 1048576
}

function pattern
{
    cleanup

    PAT="AABBCCDD"
    PAT_COUNT=12345

    # write pattern to test image file
    "$DC3DD" textpattern="$PAT" of="$TEST_IMG" log="$TEST_LOG" count="$PAT_COUNT" hash=md5,sha1 sizeprobe=on progress=on
    if [ $? -ne 0 ]
    then
        echo "Pattern: Run Failed"
        exit 3
    fi

    # check the retrieved image

    # a. file sizes
    #TODO: requires stat from coreutils
    REF_BYTES=`stat -c %s "$PAT_REF"`
    IMG_BYTES=`stat -c %s "$TEST_IMG"`

    if [ $REF_BYTES -ne $IMG_BYTES ]
    then
        echo "Pattern: Size mismatch"
        exit 3
    fi

    # b. checksums
    REF_SUM=`md5sum "$PAT_REF" | cut -f 1 -d \ `
    IMG_SUM=`md5sum "$TEST_IMG" | cut -f 1 -d \ `
    LOG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `

    if [ "$REF_SUM" != "$IMG_SUM" ]
    then
        echo "Pattern: Ref/Img Checksum mismatch"
        exit 3
    fi

    if [ "$REF_SUM" != "$LOG_SUM" ]
    then
        echo "Pattern: Ref/Log Checksum mismatch"
        exit 3
    fi

    # sector probe/count
    PAT_SECTORS="$PAT_COUNT"
    PAT_SECTORSIZE=512
    PAT_BYTES=`expr $PAT_SECTORS \* $PAT_SECTORSIZE`

    LOG_SECTORMODE=`grep sector\ size\: "$TEST_LOG" | awk '{print $4}'`
    LOG_SECTORSIZE=`grep sector\ size\: "$TEST_LOG" | awk '{print $3}'`
    LOG_SECTORS_FULL=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f1 -d+`
    LOG_SECTORS_PART=`grep sectors\ in "$TEST_LOG" | awk '{print $1}' | cut -f2 -d+`
    LOG_BYTES=`grep bytes "$TEST_LOG" | awk '{print $1}'`

    if [ "$LOG_SECTORMODE" != "(assumed)" ]
    then
        echo "Pattern: should be assuming sector size"
        exit 3
    fi

    if [ "$LOG_SECTORSIZE" != "$PAT_SECTORSIZE" ]
    then
        echo "Pattern: Wrong sector size assumed"
        exit 3
    fi

    if [ "$LOG_SECTORS_PART" != "0" ]
    then
        echo "Pattern: Partial sectors must be zero"
        exit 3
    fi

    if [ "$LOG_SECTORS_FULL" != "$PAT_SECTORS" ]
    then
        echo "Pattern: Wrong sector count"
        exit 3
    fi

    if [ "$LOG_BYTES" != "$PAT_BYTES" ]
    then
        echo "Pattern: Wrong byte count - log doesn't match size calculated from known sectors and sectorsize"
        exit 3
    fi

    if [ "$LOG_BYTES" != "$IMG_BYTES" ]
    then
        echo "Pattern: Log file bytes doesn't match output file size"
        exit 3
    fi

    if [ "$LOG_BYTES" != "$REF_BYTES" ]
    then
        echo "Pattern: Log file bytes doesn't match reference file size"
        exit 3
    fi

    MSGCOUNT=`egrep -c bytes.+copied "$TEST_LOG"`
    if [ $MSGCOUNT -ne 1 ]
    then
        echo "Pattern: Log is missing 'bytes copied' message"
        exit 3
    fi
}

function messages
{
    cleanup

    "$DC3DD" if="/dev/zero" of="/dev/null" log="$TEST_LOG" count="1000" progress=on
    if [ $? -ne 0 ]
    then
        echo "Imaging to /dev/zero test exit message failed"
        exit 6
    fi   

    LOG_VERB=`egrep "dc3dd.+?at" "$TEST_LOG" | tail -1 | awk '{print $2}'`

    if [ "$LOG_VERB" != "completed" ]
    then
        echo "Exit message verb on imaging /dev/zero success should be 'completed'"
        exit 6
    fi

    cleanup

    "$DC3DD" if="/dev/zero" of="$TEST_IMG" log="$TEST_LOG" count="1000" progress=on
    if [ $? -ne 0 ]
    then
        echo "Imaging to /dev/zero test exit message failed"
        exit 6
    fi   

    LOG_VERB=`egrep "dc3dd.+?at" "$TEST_LOG" | tail -1 | awk '{print $2}'`

    if [ "$LOG_VERB" != "completed" ]
    then
        echo "Exit message verb on imaging /dev/zero success should be 'completed'"
        exit 6
    fi

    cleanup

    "$DC3DD" if="$GOOD" of="$TEST_IMG" log="$TEST_LOG" progress=on
    if [ $? -ne 0 ]
    then
        echo "Imaging good drive to test exit message failed"
        exit 6
    fi   

    LOG_VERB=`egrep "dc3dd.+?at" "$TEST_LOG" | tail -1 | awk '{print $2}'`

    if [ "$LOG_VERB" != "completed" ]
    then
        echo "Exit message verb on imaging good drive success should be 'completed'"
        exit 6
    fi

    cleanup

    set -m
    "$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" progress=on& pid=$!;
    sleep 1; kill -s 2 %1
    wait %1
    set +m

    LOG_VERB=`egrep "dc3dd.+?at" "$TEST_LOG" | tail -1 | awk '{print $2}'`

    if [ "$LOG_VERB" != "aborted" ]
    then
        echo "Exit message verb on imaging good drive interrupt should be 'aborted'"
        exit 6
    fi
}

function rj_cat
{
    BS="$1"

    rm "$TEST_LOG"
    cat "$TEST_IMG".* | "$DC3DD" of="$TEST_IMG" log="$TEST_LOG" progress=on progresscount=1000 hash=md5 hashwindow=100M $BS
    JOIN_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `
    if [ "$ORIG_SUM" != "$JOIN_SUM" ]
    then
        echo "join regression: joining with cat & $BS, final hash mismatch"
        exit 9
    fi

    egrep "(^md5|TOTAL)" "$TEST_LOG" > "${TEST_LOG}.join_hashes"
    diff -q "${TEST_LOG}.acquire_hashes" "${TEST_LOG}.join_hashes"
    if [ $? -ne 0 ]
    then
        echo "join regression: joining with cat & $BS, piecewise hash mismatch"
        exit 9
    fi

    IMG_SUM=`md5sum "$TEST_IMG" | cut -f 1 -d \ `
    if [ "$JOIN_SUM" != "$IMG_SUM" ]
    then
        echo "join regression: joining with cat & $BS, log total hash != output file hash"
        exit 9
    fi
}

function rj_ifjoin
{
    BS="$1"

    rm "$TEST_LOG"
    "$DC3DD" ifjoin="${TEST_IMG}.000" of="$TEST_IMG" log="$TEST_LOG" sizeprobe=on progress=on progresscount=1000 hash=md5 hashwindow=100M $BS
    JOIN_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `
    if [ "$ORIG_SUM" != "$JOIN_SUM" ]
    then
        echo "join regression: joining with ifjoin & $BS, final hash mismatch"
        exit 9
    fi

    egrep "(^md5|TOTAL)" "$TEST_LOG" > "${TEST_LOG}.join_hashes"
    diff -q "${TEST_LOG}.acquire_hashes" "${TEST_LOG}.join_hashes"
    if [ $? -ne 0 ]
    then
        echo "join regression: joining with ifjoin & $BS, piecewise hash mismatch"
        exit 9
    fi

    IMG_SUM=`md5sum "$TEST_IMG" | cut -f 1 -d \ `
    if [ "$JOIN_SUM" != "$IMG_SUM" ]
    then
        echo "join regression: joining with ifjoin & $BS, log total hash != output file hash"
        exit 9
    fi
}

function regressions_join
{
    cleanup

    # get reference image
    "$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" splitformat=000 split=100M hash=md5 hashwindow=100M sizeprobe=on progress=on progresscount=1000
    if [ $? -ne 0 ]
    then
        echo "join regression: split run failed"
        exit 9
    fi

    egrep "(^md5|TOTAL)" "$TEST_LOG" > "${TEST_LOG}.acquire_hashes"

    ORIG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | cut -f 3 -d \ `

    rj_cat    "" 
    rj_cat    "bs=512" 

    rj_ifjoin "" 
    rj_ifjoin "bs=512" 
}

# test for incorrect hashwindow values because of buffer overflow
# (fixed in 6.12.3 - so far observed on Mac OS X only)
function regressions_hashwindow
{
    cleanup
        
    "$DC3DD" if="$GOOD" of="$TEST_IMG" conv=noerror,sync log="$TEST_LOG" splitformat=000 split=100M hash=md5,sha1 hashwindow=100M sizeprobe=on progress=on progresscount=1000
    if [ $? -ne 0 ]
    then
        echo "hashwindow regression: split run failed"
        exit 10
    fi

    md5sum "$TEST_IMG".0* | awk '{print $1}' > "${TEST_LOG}.ref_md5_piece_hashes"

    egrep "^md5" "$TEST_LOG" | grep -v "TOTAL" | awk '{print $4}' > "${TEST_LOG}.test_md5_piece_hashes"

    diff -q "${TEST_LOG}.ref_md5_piece_hashes" "${TEST_LOG}.test_md5_piece_hashes"
    if [ $? -ne 0 ]
    then
        echo "hashwindow regression: md5 hash window mismatch"
        exit 10
    fi

    IMG_SUM=`cat "$TEST_IMG".0* | md5sum | awk '{print $1}'`
    LOG_SUM=`grep "md5 TOTAL" "$TEST_LOG" | awk '{print $3}'`

    if [ "$IMG_SUM" != "$LOG_SUM" ]
    then
        echo "hashwindow regression: img/log md5 total mismatch"
        exit 10
    fi

    sha1sum "$TEST_IMG".0* | awk '{print $1}' > "${TEST_LOG}.ref_sha1_piece_hashes"

    egrep "^sha1" "$TEST_LOG" | grep -v "TOTAL" | awk '{print $4}' > "${TEST_LOG}.test_sha1_piece_hashes"

    diff -q "${TEST_LOG}.ref_sha1_piece_hashes" "${TEST_LOG}.test_sha1_piece_hashes"
    if [ $? -ne 0 ]
    then
        echo "hashwindow regression: sha1 hash window mismatch"
        exit 10
    fi

    IMG_SUM=`cat "$TEST_IMG".0* | sha1sum | awk '{print $1}'`
    LOG_SUM=`grep "sha1 TOTAL" "$TEST_LOG" | awk '{print $3}'`

    if [ "$IMG_SUM" != "$LOG_SUM" ]
    then
        echo "hashwindow regression: img/log sha1 total mismatch"
        exit 10
    fi
}

function regressions
{
    cleanup

    # seek crash
    "$DC3DD" if=/dev/zero of=/dev/null seek=10 count=10
    if [ $? -ne 0 ]
    then
        echo "seek operator causes failure"
        exit 7
    fi

    # ifjoin/vfjoin crash
    "$DC3DD" ifjoin=bar
    if [ $? -ne 1 ]
    then
        echo "Invalid ifjoin format not detected properly"
        exit 7
    fi

    "$DC3DD" vfjoin=bar
    if [ $? -ne 1 ]
    then
        echo "Invalid vfjoin format not detected properly"
        exit 7
    fi

    "$DC3DD" if=foo ifjoin=bar.000
    if [ $? -ne 1 ]
    then
        echo "Invalid combination of if/ifjoin not detected properly"
        exit 7
    fi

    "$DC3DD" vf=foo vfjoin=bar.000
    if [ $? -ne 1 ]
    then
        echo "Invalid combination of vf/vfjoin not detected properly"
        exit 7
    fi

    # ensure no crash when validation target is empty
    empty=`mktemp`
    "$DC3DD" if="$GOOD_REF" vf="$empty"
    if [ $? -ne 1 ]
    then
        unlink "$empty"
        echo "Crashes when verifying against empty target"
        exit 7
    fi
    unlink "$empty"
}

init

good
zeros
pattern
messages
regressions
regressions_join
regressions_hashwindow
cleanup

rm "$GOOD"

